import uuid
import json
from collections import OrderedDict
from collections.abc import Mapping
from urllib import parse

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import transaction
from django.db.models import QuerySet
from django.urls import resolve, Resolver404, get_script_prefix
from django.utils.encoding import uri_to_iri
from functools import cached_property
from rest_framework.exceptions import ValidationError
from rest_framework.fields import SkipField, empty, ReadOnlyField
from rest_framework.fields import get_error_detail
from rest_framework.serializers import HyperlinkedModelSerializer, ModelSerializer, LIST_SERIALIZER_KWARGS
from rest_framework.settings import api_settings
from rest_framework.utils import model_meta
from rest_framework.utils.field_mapping import get_nested_relation_kwargs
from rest_framework.utils.serializer_helpers import BindingDict

# Import set_value if available (DRF < 3.15), otherwise define it ourselves
try:
    from rest_framework.fields import set_value
except ImportError:
    # set_value was removed in DRF 3.15+
    # Implement it ourselves for compatibility
    def set_value(dictionary, keys, value):
        """
        Similar to Python's built in `dictionary[key] = value`,
        but takes a list of nested keys instead of a single key.

        set_value({'a': 1}, [], {'b': 2}) -> {'a': 1, 'b': 2}
        set_value({'a': 1}, ['x'], 2) -> {'a': 1, 'x': 2}
        set_value({'a': 1}, ['x', 'y'], 2) -> {'a': 1, 'x': {'y': 2}}
        """
        if not keys:
            dictionary.update(value)
            return

        for key in keys[:-1]:
            if key not in dictionary:
                dictionary[key] = {}
            dictionary = dictionary[key]

        dictionary[keys[-1]] = value

from djangoldp.fields import LDPUrlField, IdURLField
from djangoldp.models import Model
from .fields import JsonLdRelatedField, JsonLdIdentityField
from .list_serializer import ContainerSerializer
from .mixins import RDFSerializerMixin


class LDPSerializer(HyperlinkedModelSerializer, RDFSerializerMixin):
    url_field_name = "@id"
    serializer_related_field = JsonLdRelatedField
    serializer_url_field = JsonLdIdentityField
    ModelSerializer.serializer_field_mapping[LDPUrlField] = IdURLField

    # The default serializer repr ends in infinite loop. Overloading it prevents that.
    def __repr__(self):
        return self.__class__.__name__

    @cached_property
    def fields(self):
        """
        A dictionary of {field_name: field_instance}.
        """
        # `fields` is evaluated lazily. We do this to ensure that we don't
        # have issues importing modules that use ModelSerializers as fields,
        # even if Django's app-loading stage has not yet run.
        fields = BindingDict(self)

        # we allow the request object to specify a subset of fields which should be serialized
        model_fields = self.get_fields()
        req_header_accept_shape = self.context['request'].META.get('HTTP_ACCEPT_MODEL_FIELDS') if 'request' in self.context else None
        try:
            allowed_fields = list(set(json.loads(req_header_accept_shape)).intersection(model_fields.keys())) if req_header_accept_shape is not None else model_fields.keys()
        except json.decoder.JSONDecodeError:
            raise ValidationError("Please send the HTTP header Accept-Model-Fields as an array of strings")

        for key, value in model_fields.items():
            if key in allowed_fields:
                fields[key] = value

        return fields

    def get_default_field_names(self, declared_fields, model_info):
        try:
            fields = list(self.Meta.model._meta.serializer_fields)
        except AttributeError:
            fields = super().get_default_field_names(declared_fields, model_info)
        return fields + list(getattr(self.Meta, 'extra_fields', []))

    def _get_rdf_field_name(self, model_field):
        """
        :return: a string RDF type, if configured on the model_field directly
         or on a reverse relation. None, if not configured.
        """
        if getattr(model_field, "rdf_type", None) is not None:
            return model_field.rdf_type
        # If an RDF type is not configured on the foreign key directly, it may be on the reverse-key
        elif (
            hasattr(model_field, "field")
            and getattr(model_field.field, "related_rdf_type", None) is not None
        ):
            return model_field.field.related_rdf_type
        return None

    def to_representation(self, obj):
        # external Models should only be returned with rdf values
        if Model.is_external(obj):
            data = {'@id': obj.urlid}
            return self.serialize_rdf_fields(obj, data)

        data = super().to_representation(obj)

        slug_field = Model.slug_field(obj)
        for field in data:
            if isinstance(data[field], dict) and '@id' in data[field]:
                data[field]['@id'] = data[field]['@id'].format(Model.container_id(obj), str(getattr(obj, slug_field)))
        # prioritise urlid field over generated @id
        if 'urlid' in data and data['urlid'] is not None:
            data['@id'] = data.pop('urlid')['@id']
        if not '@id' in data:
            data['@id'] = '{}{}'.format(settings.SITE_URL, Model.resource(obj))

        # Django Rest Framework will by default serialize fields with the field name.
        # LDPFields may have configured an RDF type which is required for valid serialization.
        # This is handled after serialization to avoid overriding the basic serialization of DRF.
        model = self.Meta.model
        for field in self._readable_fields:
            try:
                model_field = model._meta.get_field(field.source)
                if (
                    model_field is not None
                    and field.field_name in data
                    and self._get_rdf_field_name(model_field) is not None
                ):
                    data[self._get_rdf_field_name(model_field)] = data[field.field_name]
                    data.pop(field.field_name)
            except FieldDoesNotExist:
                pass

        data = self.serialize_rdf_fields(obj, data, include_context=True)
        data = self.add_permissions(data, self.context['request'].user, type(obj), obj=obj)
        return data

    def build_property_field(self, field_name, model_class):
        class JSonLDPropertyField(ReadOnlyField):
            def to_representation(self, instance):
                from djangoldp.views.ldp_viewset import LDPViewSet
                if isinstance(instance, QuerySet):
                    model = instance.model
                elif isinstance(instance, Model):
                    model = type(instance)
                else:
                    return instance
                depth = max(getattr(self.parent.Meta, "depth", 0) - 1, 0)
                fields = ["@id"] if depth==0 else getattr(model._meta, 'serializer_fields', [])

                serializer_generator = LDPViewSet(model=model, fields=fields, depth=depth,
                                                    lookup_field=getattr(model._meta, 'lookup_field', 'pk'),
                                                    permission_classes=getattr(model._meta, 'permission_classes', []),
                                                    nested_fields=getattr(model._meta, 'nested_fields', []))
                serializer = serializer_generator.get_serializer_class()(context=self.parent.context)

                if isinstance(instance, QuerySet):
                    id = '{}{}{}/'.format(settings.SITE_URL, '{}{}/', self.source)
                    children = [serializer.to_representation(item) for item in instance]
                    return self.parent.serialize_container(children, id, self.parent.context['request'].user, model)
                else:
                    return serializer.to_representation(instance)

        return JSonLDPropertyField, {}

    def handle_value_object(self, value):
        '''
        In JSON-LD value-objects can be passed in, which store some context on the field passed. By overriding this
        function you can react to this context on a field without overriding build_standard_field
        '''
        return value['@value']

    def build_standard_field(self, field_name, model_field):
        class JSonLDStandardField:
            parent_view_name = None

            def __init__(self, **kwargs):
                self.parent_view_name = kwargs.pop('parent_view_name')
                super().__init__(**kwargs)

            def get_value(self, dictionary):
                if self.field_name == 'urlid':
                    self.field_name = '@id'
                try:
                    object_list = dictionary["@graph"]
                    if self.parent.instance is None:
                        obj = next(filter(
                            lambda o: not self.parent.url_field_name in o or "./" in o[self.parent.url_field_name],
                            object_list))
                        value = super().get_value(obj)
                    else:
                        resource_id = Model.resource_id(self.parent.instance)
                        obj = next(
                            filter(lambda o: resource_id.lstrip('/') in o[self.parent.url_field_name], object_list))
                        value = super().get_value(obj)
                except KeyError:
                    value = super().get_value(dictionary)

                if self.field_name == '@id' and value == './':
                    self.field_name = 'urlid'
                    return None

                if isinstance(value, dict) and '@value' in value:
                    value = self.parent.handle_value_object(value)

                return self.manage_empty(value)

            def manage_empty(self, value):
                if value == '' and self.allow_null:
                    # If the field is blank, and null is a valid value then
                    # determine if we should use null instead.
                    return '' if getattr(self, 'allow_blank', False) else None
                elif value == '' and not self.required:
                    # If the field is blank, and emptiness is valid then
                    # determine if we should use emptiness instead.
                    return '' if getattr(self, 'allow_blank', False) else empty
                return value

        field_class, field_kwargs = super().build_standard_field(field_name, model_field)
        field_kwargs['parent_view_name'] = '{}-list'.format(model_field.model._meta.object_name.lower())
        return type(field_class.__name__ + 'Valued', (JSonLDStandardField, field_class), {}), field_kwargs

    def build_nested_field(self, field_name, relation_info, nested_depth):
        class NestedLDPSerializer(self.__class__):
            class Meta:
                model = relation_info.related_model
                depth = nested_depth - 1
                try:
                    fields = ['@id'] + list(model._meta.serializer_fields)
                except AttributeError:
                    fields = '__all__'

            def to_internal_value(self, data):
                if data == '':
                    return ''
                if self.url_field_name in data:
                    if not isinstance(data, Mapping):
                        message = self.error_messages['invalid'].format(
                            datatype=type(data).__name__
                        )
                        raise ValidationError({
                            api_settings.NON_FIELD_ERRORS_KEY: [message]
                        }, code='invalid')

                    ret = OrderedDict()
                    errors = OrderedDict()

                    # validate fields passed in the data
                    for field in self._writable_fields:
                        # Consider only fields which are present in data, but try to find alternative keys for that field
                        # in the data, first.
                        if field.field_name not in data:
                            try:
                                model_field = self.Meta.model._meta.get_field(field.source)
                                rdf_field_name = self._get_rdf_field_name(model_field)
                                if (
                                    model_field is not None
                                    and field.field_name not in data
                                    and rdf_field_name is not None
                                    and rdf_field_name in data
                                ):
                                    data[field.field_name] = data[rdf_field_name]
                                    data.pop(rdf_field_name)
                                else:
                                    continue
                            except FieldDoesNotExist:
                                continue

                        validate_method = getattr(self, 'validate_' + field.field_name, None)
                        primitive_value = field.get_value(data)
                        try:
                            validated_value = field.run_validation(primitive_value)
                            if validate_method is not None:
                                validated_value = validate_method(validated_value)
                        except ValidationError as exc:
                            errors[field.field_name] = exc.detail
                        except DjangoValidationError as exc:
                            errors[field.field_name] = get_error_detail(exc)
                        except SkipField:
                            pass
                        else:
                            set_value(ret, field.source_attrs, validated_value)

                    if errors:
                        raise ValidationError(errors)

                    # if it's a local resource - use the path to resolve the slug_field on the model
                    uri = data[self.url_field_name]
                    if not Model.is_external(uri):
                        http_prefix = uri.startswith(('http:', 'https:'))

                        if http_prefix:
                            uri = parse.urlparse(uri).path
                            prefix = get_script_prefix()
                            if uri.startswith(prefix):
                                uri = '/' + uri[len(prefix):]

                        try:
                            match = resolve(uri_to_iri(uri))
                            slug_field = Model.slug_field(self.__class__.Meta.model)
                            ret[slug_field] = match.kwargs[slug_field]
                        except Resolver404:
                            pass

                    if 'urlid' in data:
                        ret['urlid'] = data['urlid']

                else:
                    ret = super().to_internal_value(data)

                # copy url_field_name value to urlid, if necessary
                if self.url_field_name in data and not 'urlid' in data and data[self.url_field_name].startswith('http'):
                    ret['urlid'] = data[self.url_field_name]

                return ret

        kwargs = get_nested_relation_kwargs(relation_info)
        kwargs['read_only'] = False
        kwargs['required'] = False
        return NestedLDPSerializer, kwargs

    @classmethod
    def many_init(cls, *args, **kwargs):
        allow_empty = kwargs.pop('allow_empty', None)
        child_serializer = cls(*args, **kwargs)
        list_kwargs = {
            'child': child_serializer,
        }
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty
        list_kwargs.update({
            key: value for key, value in kwargs.items()
            if key in LIST_SERIALIZER_KWARGS
        })
        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(meta, 'list_serializer_class', ContainerSerializer)
        serializer = list_serializer_class(*args, **list_kwargs)

        # if the child serializer has disabled the cache, really it means disable it on the container
        if hasattr(child_serializer, 'with_cache'):
            serializer.with_cache = child_serializer.with_cache

        return serializer

    def to_internal_value(self, data):
        # TODO: This hack is needed because external users don't pass validation.
        # Objects require all fields to be optional to be created as external, and username is required.
        is_user_and_external = self.Meta.model is get_user_model() and '@id' in data and Model.is_external(data['@id'])
        if is_user_and_external:
            data['username'] = 'external'

        # Django Rest Framework will by default parse fields using the field name.
        # The user may give us JSON-LD, which will have been compacted by the JSON-LD parser, but which may still have
        # a namespace associated to it.
        # Currently this data may not always find the correct field, but we make a best-effort by matching
        # the namespace of the data given to an rdf_type configured on that field.
        # TODO: This is not a robust solution but a temporary mitigation.
        # The parser may compact the field to a different namespace or it may not be able to compact the field at all.
        for field in self._writable_fields:
            try:
                model_field = self.Meta.model._meta.get_field(field.source)
                rdf_field_name = self._get_rdf_field_name(model_field)
                if (
                    model_field is not None
                    and field.field_name not in data
                    and rdf_field_name is not None
                    and rdf_field_name in data
                ):
                    data[field.field_name] = data[rdf_field_name]
                    data.pop(rdf_field_name)
            except FieldDoesNotExist:
                pass

        ret = super().to_internal_value(data)
        if is_user_and_external:
            ret['urlid'] = data['@id']
            ret.pop('username')
        return ret

    def get_value(self, dictionary):
        '''overrides get_value to handle @graph key'''
        try:
            object_list = dictionary["@graph"]
            if self.parent.instance is None:
                obj = next(filter(
                    lambda o: not hasattr(o, self.parent.url_field_name) or "./" in o[self.url_field_name],
                    object_list))
            else:
                container_id = Model.container_id(self.parent.instance)
                obj = next(filter(lambda o: container_id.lstrip('/') in o[self.url_field_name], object_list))
            item = super().get_value(obj)
            full_item = None
            if item is empty:
                return empty
            try:
                full_item = next(
                    filter(lambda o: self.url_field_name in o and (item[self.url_field_name] == o[self.url_field_name]),
                           object_list))
            except StopIteration:
                pass
            if full_item is None:
                return item
            else:
                return full_item

        except KeyError:
            return super().get_value(dictionary)

    def create(self, validated_data):
        with transaction.atomic():
            instance = self.internal_create(validated_data, model=self.Meta.model)
            self.attach_related_object(instance, validated_data)

        return instance

    def attach_related_object(self, instance, validated_data):
        '''adds m2m relations included in validated_data to the instance'''
        model_class = self.Meta.model

        info = model_meta.get_field_info(model_class)
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and relation_info.reverse and not field_name is None:
                rel = getattr(instance._meta.model, field_name).rel
                if rel.name in validated_data:
                    related = validated_data[rel.name]
                    getattr(instance, field_name).add(related)

    def internal_create(self, validated_data, model):
        validated_data = self.resolve_fk_instances(model, validated_data, True)

        # build tuples list of nested_field keys and their values. All list values are considered nested fields
        nested_fields = []
        nested_list_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
        for field_name in nested_list_fields_name:
            nested_fields.append((field_name, validated_data.pop(field_name)))

        info = model_meta.get_field_info(model)
        many_to_many = []
        one_to_one = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and relation_info.reverse and (
                    field_name in validated_data) and not field_name is None:
                many_to_many.append((field_name, validated_data.pop(field_name)))
            elif relation_info.reverse and (field_name in validated_data) and not field_name is None:
                one_to_one[field_name] = validated_data.pop(field_name)
        validated_data = self.remove_empty_value(validated_data)

        if model is get_user_model() and not 'username' in validated_data:
            validated_data['username'] = str(uuid.uuid4())
        instance = model.objects.create(**validated_data)

        for field_name, value in many_to_many:
            validated_data[field_name] = value

        if one_to_one:
             for field_name, value in one_to_one.items():
                 setattr(instance, field_name, value)
                 value.save()

        self.save_or_update_nested_list(instance, nested_fields)

        return instance

    def remove_empty_value(self, validated_data):
        '''sets any empty strings in the validated_data to None'''
        for attr, value in validated_data.items():
            if value == '':
                validated_data[attr] = None
        return validated_data

    def update(self, instance, validated_data):
        model = self.Meta.model
        nested_fields = []
        nested_fields_name = list(filter(lambda key: isinstance(validated_data[key], list), validated_data))
        for field_name in nested_fields_name:
            nested_fields.append((field_name, validated_data.pop(field_name)))

        for attr, value in validated_data.items():
            if isinstance(value, dict):
                value = self.update_dict_value(attr, instance, value)
            if value == '' and not isinstance(getattr(instance, attr), str):
                setattr(instance, attr, None)
            else:
                setattr(instance, attr, value)

        self.save_or_update_nested_list(instance, nested_fields)
        instance.save()

        return instance

    def resolve_fk_instances(self, model, validated_data, create=False):
        '''
        iterates over every dict object in validated_data and resolves them into instances (get or create)
        :param model: the model being operated on
        :param validated_data: the data passed to the serializer
        :param create: set to True, foreign keys will be created if they do not exist
        '''
        nested_fk_fields_name = list(filter(lambda key: isinstance(validated_data[key], dict), validated_data))
        for field_name in nested_fk_fields_name:
            field_dict = validated_data[field_name]
            field_model = model._meta.get_field(field_name).related_model

            slug_field = Model.slug_field(field_model)
            sub_inst = None
            if 'urlid' in field_dict:
                # has urlid and is a local resource
                model, sub_inst = self.get_inst_by_urlid(field_dict, field_model, model, slug_field, sub_inst)
            # try slug field, assuming that this is a local resource
            elif slug_field in field_dict:
                kwargs = {slug_field: field_dict[slug_field]}
                sub_inst = field_model.objects.get(**kwargs)
            if sub_inst is None:
                if create:
                    sub_inst = self.internal_create(field_dict, field_model)
                else:
                    continue

            validated_data[field_name] = sub_inst
        return validated_data

    def get_inst_by_urlid(self, field_dict, field_model, model, slug_field, sub_inst):
        if parse.urlparse(settings.BASE_URL).netloc == parse.urlparse(field_dict['urlid']).netloc:
            # try slug field if it exists
            if slug_field in field_dict:
                kwargs = {slug_field: field_dict[slug_field]}
                sub_inst = field_model.objects.get(**kwargs)
            else:
                model, sub_inst = Model.resolve(field_dict['urlid'])
        # remote resource - get backlinked copy
        elif hasattr(field_model, 'urlid'):
            sub_inst = Model.get_or_create_external(field_model, field_dict['urlid'])
        return model, sub_inst

    def update_dict_value(self, attr, instance, value):
        info = model_meta.get_field_info(instance)
        relation_info = info.relations.get(attr)
        slug_field = Model.slug_field(relation_info.related_model)

        if slug_field in value:
            value = self.update_dict_value_when_id_is_provided(attr, instance, relation_info, slug_field, value)
        else:
            if 'urlid' in value:
                if parse.urlparse(settings.BASE_URL).netloc == parse.urlparse(value['urlid']).netloc:
                    model, oldObj = Model.resolve(value['urlid'])
                    value = self.update(instance=oldObj, validated_data=value)
                elif hasattr(relation_info.related_model, 'urlid'):
                    value = Model.get_or_create_external(relation_info.related_model, value['urlid'])
            else:
                value = self.update_dict_value_without_slug_field(attr, instance, relation_info, value)
        return value

    def update_dict_value_without_slug_field(self, attr, instance, relation_info, value):
        if relation_info.to_many:
            value = self.internal_create(validated_data=value, model=relation_info.related_model)
        else:
            rel = instance._meta.get_field(attr)
            reverse_attr_name = rel.remote_field.name
            many = rel.one_to_many or rel.many_to_many
            if many:
                value[reverse_attr_name] = [instance]
                oldObj = rel.model.object.get(id=value['urlid'])
            else:
                value[reverse_attr_name] = instance
                oldObj = getattr(instance, attr, None)

            if oldObj is None:
                value = self.internal_create(validated_data=value, model=relation_info.related_model)
            else:
                value = self.update(instance=oldObj, validated_data=value)
        return value

    def update_dict_value_when_id_is_provided(self, attr, instance, relation_info, slug_field, value):
        kwargs = {slug_field: value[slug_field]}
        if relation_info.to_many:
            manager = getattr(instance, attr)
            oldObj = manager._meta.model.objects.get(**kwargs)
        else:
            related_model = relation_info.related_model
            oldObj = related_model.objects.get(**kwargs)

        value = self.update(instance=oldObj, validated_data=value)
        return value

    def save_or_update_nested_list(self, instance, nested_fields):
        for (field_name, data) in nested_fields:
            manager = getattr(instance, field_name)
            field_model = manager.model
            slug_field = Model.slug_field(field_model)
            try:
                item_pk_to_keep = [obj_dict[slug_field] for obj_dict in data if slug_field in obj_dict]
            except TypeError:
                item_pk_to_keep = [getattr(obj, slug_field) for obj in data if hasattr(obj, slug_field)]

            if hasattr(manager, 'through'):
                manager.clear()
            else:
                manager.exclude(pk__in=item_pk_to_keep).delete()

            for item in data:
                if isinstance(item, Model):
                    item.save()
                    saved_item = item
                elif slug_field in item:
                    kwargs = {slug_field: item[slug_field]}
                    saved_item = self.get_or_create(field_model, item, kwargs)
                elif 'urlid' in item:
                    # has urlid and is a local resource
                    if not Model.is_external(item['urlid']):
                        model, old_obj = Model.resolve(item['urlid'])
                        if old_obj is not None:
                            saved_item = self.update(instance=old_obj, validated_data=item)
                        else:
                            saved_item = self.internal_create(validated_data=item, model=field_model)
                    # has urlid and is external resource
                    elif hasattr(field_model, 'urlid'):
                        kwargs = {'urlid': item['urlid']}
                        saved_item = self.get_or_create(field_model, item, kwargs)
                else:
                    rel = getattr(instance._meta.model, field_name).rel
                    try:
                        if rel.related_model == manager.model:
                            reverse_id = rel.remote_field.attname
                            item[reverse_id] = instance.pk
                    except AttributeError:
                        pass
                    saved_item = self.internal_create(validated_data=item, model=manager.model)

                if hasattr(manager, 'through') and manager.through._meta.auto_created:
                    # First remove to avoid duplicates
                    manager.remove(saved_item)
                    manager.add(saved_item)

    def get_or_create(self, field_model, item, kwargs):
        try:
            old_obj = field_model.objects.get(**kwargs)
            saved_item = self.update(instance=old_obj, validated_data=item)
        except field_model.DoesNotExist:
            saved_item = self.internal_create(validated_data=item, model=field_model)
        return saved_item



from datetime import date
from django.conf import settings
from django.contrib.auth.models import AbstractUser, Group
from django.db import models

from djangoldp import fields
from djangoldp.models import Model, DynamicNestedField
from djangoldp.permissions import ACLPermissions, AuthenticatedOnly, ReadOnly, \
    ReadAndCreate, AnonymousReadOnly, OwnerPermissions, InheritPermissions

from .permissions import Only2WordsForToto, ReadOnlyStartsWithA


class User(AbstractUser, Model):
    class Meta(AbstractUser.Meta, Model.Meta):
        depth = 0
        empty_containers = []
        ordering = ['pk']
        serializer_fields = ['@id', 'username', 'first_name', 'last_name', 'email', 'userprofile',
                             'conversation_set','groups', 'projects', 'owned_circles', 'affiliates']
        permission_classes = [ReadAndCreate|OwnerPermissions]
        rdf_type = 'foaf:user'
        nested_fields = ['owned_circles', 'affiliates']


class Skill(Model):
    title = models.CharField(max_length=255, blank=True, null=True)
    obligatoire = models.CharField(max_length=255)
    slug = models.SlugField(blank=True, null=True, unique=True)
    date = models.DateTimeField(auto_now_add=True, blank=True)

    def recent_jobs(self):
        return self.joboffer_set.filter(date__gte=date.today())

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]
        serializer_fields = ["@id", "title", "recent_jobs", "slug", "obligatoire"]
        nested_fields = ['joboffer_set']
        lookup_field = 'slug'
        rdf_type = 'hd:skill'


class JobOffer(Model):
    title = models.CharField(max_length=255, null=True)
    skills = models.ManyToManyField(Skill, blank=True)
    slug = models.SlugField(blank=True, null=True, unique=True)
    date = models.DateTimeField(auto_now_add=True, blank=True)

    def recent_skills(self):
        return self.skills.filter(date__gte=date.today())

    def some_skill(self):
        return self.skills.all().first()

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly, ReadOnly|OwnerPermissions]
        serializer_fields = ["@id", "title", "skills", "recent_skills", "resources", "slug", "some_skill", "urlid"]
        nested_fields = ['skills', 'resources', 'recent_skills']
        container_path = "job-offers/"
        lookup_field = 'slug'
        rdf_type = 'hd:joboffer'

JobOffer.recent_skills.field = DynamicNestedField(Skill, 'recent_skills')


class Conversation(models.Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    author_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING)
    peer_user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="peers_conv",
                                  on_delete=models.DO_NOTHING)
    observers = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True, related_name='observed_conversations')

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]
        nested_fields=["message_set", "observers"]
        owner_field = 'author_user'


class Resource(Model):
    joboffers = models.ManyToManyField(JobOffer, blank=True, related_name='resources')
    description = models.CharField(max_length=255)

    class Meta(Model.Meta):
        ordering = ['pk']
        serializer_fields = ["@id", "joboffers"]
        nested_fields = ['joboffers']
        depth = 1
        rdf_type = 'hd:Resource'


# a resource in which only the owner has permissions (for testing owner permissions)
class OwnedResource(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="owned_resources",
                             on_delete=models.CASCADE)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [OwnerPermissions]
        owner_field = 'user'
        serializer_fields = ['@id', 'description', 'user']
        nested_fields = ['owned_resources']
        depth = 1


class OwnedResourceVariant(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="owned_variant_resources",
                             on_delete=models.CASCADE)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadOnly|OwnerPermissions]
        owner_field = 'user'
        serializer_fields = ['@id', 'description', 'user']
        depth = 1


class OwnedResourceNestedOwnership(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    parent = models.ForeignKey(OwnedResource, blank=True, null=True, related_name="owned_resources",
                               on_delete=models.CASCADE)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [OwnerPermissions]
        owner_field = 'parent__user'
        serializer_fields = ['@id', 'description', 'parent']
        nested_fields = ['owned_resources']
        depth = 1


class OwnedResourceTwiceNestedOwnership(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    parent = models.ForeignKey(OwnedResourceNestedOwnership, blank=True, null=True, related_name="owned_resources",
                               on_delete=models.CASCADE)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [OwnerPermissions]
        owner_field = 'parent__parent__user'
        serializer_fields = ['@id', 'description', 'parent']
        depth = 1


class UserProfile(Model):
    description = models.CharField(max_length=255, blank=True, null=True)
    user = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='userprofile', on_delete=models.CASCADE)
    slug = models.SlugField(blank=True, null=True, unique=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AuthenticatedOnly,ReadOnly|OwnerPermissions]
        owner_field = 'user'
        lookup_field = 'slug'
        serializer_fields = ['@id', 'description', 'settings', 'user']
        depth = 1


class NotificationSetting(Model):
    user = models.OneToOneField(UserProfile, on_delete=models.CASCADE, related_name="settings")
    receiveMail = models.BooleanField(default=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadAndCreate|OwnerPermissions]


class Message(models.Model):
    text = models.CharField(max_length=255, blank=True, null=True)
    conversation = models.ForeignKey(Conversation, on_delete=models.DO_NOTHING)
    author_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]


class Dummy(models.Model):
    some = models.CharField(max_length=255, blank=True, null=True)
    slug = models.SlugField(blank=True, null=True, unique=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]


class LDPDummy(Model):
    some = models.CharField(max_length=255, blank=True, null=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]
        nested_fields = ['anons']


# model used in django-guardian permission tests (no permission to anyone except suuperusers)
class PermissionlessDummy(Model):
    some = models.CharField(max_length=255, blank=True, null=True)
    slug = models.SlugField(blank=True, null=True, unique=True)
    parent = models.ForeignKey(LDPDummy, on_delete=models.DO_NOTHING, related_name="anons", blank=True, null=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ACLPermissions]
        lookup_field='slug'
        permissions = (('custom_permission_permissionlessdummy', 'Custom Permission'),)


class Post(Model):
    content = models.CharField(max_length=255)
    author = models.ForeignKey(User, blank=True, null=True, on_delete=models.SET_NULL)
    peer_user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, related_name="peers_post",
                                  on_delete=models.SET_NULL)
    # Timestamp fields (created_at, updated_at) are inherited from Model base class

    class Meta(Model.Meta):
        ordering = ['pk']
        auto_author = 'author'
        rdf_type = 'hd:post'

class AnonymousReadOnlyPost(Model):
    content = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AnonymousReadOnly]
class AuthenticatedOnlyPost(Model):
    content = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [AuthenticatedOnly]
class ReadOnlyPost(Model):
    content = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadOnly]
class ReadAndCreatePost(Model):
    content = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadAndCreate]

class ANDPermissionsDummy(Model):
    title = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadOnlyStartsWithA&Only2WordsForToto]
class ORPermissionsDummy(Model):
    title = models.CharField(max_length=255)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ReadOnlyStartsWithA|Only2WordsForToto]


class Invoice(Model):
    title = models.CharField(max_length=255, blank=True, null=True)
    date = models.DateField(blank=True, null=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        depth = 2
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]
        nested_fields = ['batches']


class Circle(Model):
    name = models.CharField(max_length=255, blank=True)
    description = models.CharField(max_length=255, blank=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="owned_circles", on_delete=models.DO_NOTHING, null=True, blank=True)
    members = models.OneToOneField(Group, related_name="circle", on_delete=models.SET_NULL, null=True, blank=True)
    admins = models.OneToOneField(Group, related_name="admin_circle", on_delete=models.SET_NULL, null=True, blank=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        auto_author = 'owner'
        depth = 1
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions|ACLPermissions]
        permission_roles = {
            'members': {'perms': ['view'], 'add_author': True},
            'admins': {'perms': ['view', 'change', 'control'], 'add_author': True},
        }
        serializer_fields = ['@id', 'name', 'description', 'members', 'owner', 'space']
        rdf_type = 'hd:circle'

Group._meta.inherit_permissions += ['circle','admin_circle']
Group._meta.serializer_fields += ['circle', 'admin_circle']

class RestrictedCircle(Model):
    name = models.CharField(max_length=255, blank=True)
    description = models.CharField(max_length=255, blank=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="owned_restrictedcircles", on_delete=models.DO_NOTHING, null=True, blank=True)
    members = models.ForeignKey(Group, related_name="restrictedcircles", on_delete=models.SET_NULL, null=True, blank=True)
    admins = models.ForeignKey(Group, related_name="admin_restrictedcircles", on_delete=models.SET_NULL, null=True, blank=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        auto_author = 'owner'
        permission_classes = [ACLPermissions]
        permission_roles = {
            'members': {'perms': ['view'], 'add_author': True},
            'admins': {'perms': ['view', 'change', 'control'], 'add_author': True},
        }
        rdf_type = 'hd:circle'
class RestrictedResource(Model):
    content = models.CharField(max_length=255, blank=True)
    circle = models.ForeignKey(RestrictedCircle, on_delete=models.CASCADE)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [InheritPermissions]
        inherit_permissions = ['circle']

class DoubleInheritModel(Model):
    content = models.CharField(max_length=255, blank=True)
    ro_ancestor = models.ForeignKey(ReadOnlyPost, on_delete=models.CASCADE, null=True, blank=True)
    circle = models.ForeignKey(RestrictedCircle, on_delete=models.CASCADE, null=True, blank=True)
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [InheritPermissions]
        inherit_permissions = ['circle', 'ro_ancestor']

class Space(Model):
    name = models.CharField(max_length=255, blank=True)
    circle = models.OneToOneField(to=Circle, null=True, blank=True, on_delete=models.CASCADE, related_name='space')

    class Meta(Model.Meta):
        ordering = ['pk']


class Batch(Model):
    invoice = models.ForeignKey(Invoice, on_delete=models.CASCADE, related_name='batches')
    title = models.CharField(max_length=255, blank=True, null=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        serializer_fields = ['@id', 'title', 'invoice', 'tasks']
        permission_classes = [ReadAndCreate|OwnerPermissions]
        depth = 1
        rdf_type = 'hd:batch'


class Task(models.Model):
    batch = models.ForeignKey(Batch, on_delete=models.CASCADE, related_name='tasks')
    title = models.CharField(max_length=255)

    class Meta(Model.Meta):
        ordering = ['pk']
        serializer_fields = ['@id', 'title', 'batch']
        permission_classes = [AnonymousReadOnly,ReadAndCreate|OwnerPermissions]


class ModelTask(Model, Task):
    class Meta(Model.Meta):
        ordering = ['pk']

STATUS_CHOICES = [
    ('Public', 'Public'),
    ('Private', 'Private'),
    ('Archived', 'Archived'),
]

class Project(Model):
    description = models.CharField(max_length=255, null=True, blank=False)
    status = models.CharField(max_length=8, choices=STATUS_CHOICES, default='Private', null=True, blank=True)
    members = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True, related_name='projects')

    class Meta(Model.Meta):
        ordering = ['pk']
        rdf_type = 'hd:project'
        nested_fields = ['members']


class DateModel(Model):
    excluded = models.CharField(max_length=255, null=True, default='test')
    value = models.DateField()

    class Meta(Model.Meta):
        ordering = ['pk']
        rdf_type = "hd:date"
        serializer_fields_exclude = ['excluded']


class DateChild(Model):
    parent = models.ForeignKey(DateModel, on_delete=models.CASCADE, related_name='children')

    class Meta(Model.Meta):
        ordering = ['pk']
        rdf_type = 'hd:datechild'


class MyAbstractModel(Model):
    defaultsomething = models.CharField(max_length=255, blank=True)

    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ACLPermissions]
        abstract = True
        rdf_type = "wow:defaultrdftype"


class NoSuperUsersAllowedModel(Model):
    class Meta(Model.Meta):
        ordering = ['pk']
        permission_classes = [ACLPermissions]


class Enterprise(Model):
    name = fields.TextField(rdf_type="dfc-b:name", blank=True, null=True)
    VATstatus = fields.BooleanField(
        rdf_type="dfc-b:VATStatus", default=False, null=True
    )
    affiliated_to = fields.ManyToManyField(
        User,
        rdf_type="dfc-b:affiliatedTo",
        related_rdf_type="dfc-b:affiliates",
        related_name="affiliates",
        blank=True,
    )

    class Meta(Model.Meta):
        rdf_type = "dfc-b:Enterprise"



from django.db.models import options

__version__ = "5.0.7"

options.DEFAULT_NAMES += (
    "lookup_field",
    "rdf_type",
    "rdf_context",
    "auto_author",
    "owner_field",
    "owner_urlid_field",
    "view_set",
    "container_path",
    "permission_classes",
    "serializer_fields",
    "serializer_fields_exclude",
    "empty_containers",
    "nested_fields",
    "depth",
    "permission_roles",
    "include_inner_perms",
    "inherit_permissions",
    "public_field",
    "static_version",
    "static_params",
    "active_field",
    "disable_url",
)



import json

from django.test import TestCase
from rest_framework.test import APIRequestFactory, APITestCase
from djangoldp.views.type_index import PublicTypeIndexView


class PublicTypeIndexViewTests(APITestCase):
    def setUp(self):
        self.factory = APIRequestFactory()

    def test_public_type_index_has_graph(self):
        request = self.factory.get('/profile/publicTypeIndex')
        view = PublicTypeIndexView.as_view()
        response = view(request)
        response.render()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'application/ld+json')

        data = json.loads(response.content)
        self.assertIn('@context', data)
        self.assertIn('@graph', data)
        self.assertIsInstance(data['@graph'], list)

        first = data['@graph'][0]
        self.assertIn('@id', first)
        self.assertIn('@type', first)


from copy import copy

from django.conf import settings
from django.db.models import QuerySet
from rest_framework.fields import empty

from djangoldp.models import Model
from djangoldp.permissions import DEFAULT_DJANGOLDP_PERMISSIONS

from .cache import GLOBAL_SERIALIZER_CACHE


class RDFSerializerMixin:
    def add_permissions(self, data, user, model, obj=None):
        '''takes a set or list of permissions and returns them in the JSON-LD format'''
        if self.parent and not settings.LDP_INCLUDE_INNER_PERMS and not getattr(model._meta, 'include_inner_perms', False):  # Don't serialize permissions on nested objects
            return data

        if user.is_superuser:
            data['permissions'] = getattr(settings, 'DJANGOLDP_PERMISSIONS', DEFAULT_DJANGOLDP_PERMISSIONS)
            return data

        permission_classes = getattr(model._meta, 'permission_classes', [])
        if not permission_classes:
            return data
        # The permissions must be given by all permission classes to be granted
        permissions = set.intersection(*[permission().get_permissions(user, model, obj) for permission in permission_classes])
        # Don't grant delete permissions on containers
        if not obj and 'delete' in permissions:
            permissions.remove('delete')
        data['permissions'] = permissions
        return data

    def serialize_rdf_fields(self, obj, data, include_context=False):
        '''adds the @type and the @context to the data'''
        rdf_type = getattr(obj._meta, 'rdf_type', None)
        rdf_context = getattr(obj._meta, 'rdf_context', None)
        if rdf_type:
            data['@type'] = rdf_type
        if include_context and rdf_context:
            data['@context'] = rdf_context
        return data

    def serialize_container(self, data, id, user, model, obj=None):
        '''turns a list into a container representation'''
        return self.add_permissions({'@id': id, '@type': 'ldp:Container', 'ldp:contains': data}, user, model, obj)


class LDListMixin(RDFSerializerMixin):
    '''A Mixin for serializing containers into JSONLD format'''
    child_attr = 'child'
    with_cache = getattr(settings, 'SERIALIZER_CACHE', True)

    def get_child(self):
        return getattr(self, self.child_attr)

    # converts primitive data representation to the representation used within our application
    def to_internal_value(self, data):
        try:
            # if this is a container, the data will be stored in ldp:contains
            data = data['ldp:contains']
        except (TypeError, KeyError):
            pass

        if len(data) == 0:
            return []
        if isinstance(data, dict):
            data = [data]
        if isinstance(data, str) and data.startswith("http"):
            data = [{'@id': data}]

        return [self.get_child().to_internal_value(item) for item in data]

    def filter_queryset(self, queryset, child_model):
        '''Applies the permission of the child model to the child queryset'''
        view = copy(self.context['view'])
        view.model = child_model
        filter_backends = list({perm_class().get_filter_backend(child_model) for perm_class in
                                getattr(child_model._meta, 'permission_classes', []) if hasattr(perm_class(), 'get_filter_backend')})
        for backend in filter_backends:
            if backend:
                queryset = backend().filter_queryset(self.context['request'], queryset, view)
        return queryset

    def compute_id(self, value):
        '''generates the @id of the container'''
        if not hasattr(self, 'parent_instance'):
            # This is a container
            return f"{settings.BASE_URL}{self.context['request'].path}"
        return f"{settings.BASE_URL}{Model.resource_id(self.parent_instance)}{self.field_name}/"

    def get_attribute(self, instance):
        # save the parent object for nested field url
        self.parent_instance = instance
        return super().get_attribute(instance)

    def check_cache(self, value, id, model, cache_vary):
        '''Auxiliary function to avoid code duplication - checks cache and returns from it if it has entry'''
        parent_meta = getattr(self.get_child(), 'Meta', getattr(self.parent, 'Meta', None))
        depth = max(getattr(parent_meta, "depth", 0), 0) if parent_meta else 1

        if depth:
            # if the depth is greater than 0, we don't hit the cache, because a nested container might be outdated
            # this Mixin may not have access to the depth of the parent serializer, e.g. if it's a ManyRelatedField
            # in these cases we assume the depth is 0 and so we hit the cache
            return None
        cache_key = getattr(model._meta, 'label', None)

        if self.with_cache and GLOBAL_SERIALIZER_CACHE.has(cache_key, id, cache_vary):
            cache_value = GLOBAL_SERIALIZER_CACHE.get(cache_key, id, cache_vary)
            # this check is to handle the situation where the cache has been invalidated by something we don't check
            # namely if my permissions are upgraded then I may have access to view more objects
            cache_under_value = cache_value['ldp:contains'] if 'ldp:contains' in cache_value else cache_value
            if not hasattr(cache_under_value, '__len__') or not hasattr(value, '__len__') or (len(cache_under_value) == len(value)):
                return cache_value
        return False

    def to_representation(self, value):
        '''
        Converts internal representation to primitive data representation
        Filters objects out which I don't have permission to view
        Permission on container :
         - Can Add if add permission on contained object's type
         - Can view the container is view permission on container model : container obj are filtered by view permission
        '''
        try:
            child_model = self.get_child().Meta.model
        except AttributeError:
            child_model = value.model
        user = self.context['request'].user
        id = self.compute_id(value)

        is_container = True
        if getattr(self, 'parent', None):  # If we're in a nested container
            if isinstance(value, QuerySet) and getattr(self, 'parent', None):
                value = self.filter_queryset(value, child_model)

            if getattr(self, 'field_name', None) is not None:
                if self.field_name in getattr(self.parent.Meta.model._meta, 'empty_containers', []):
                    return {'@id': id}
                if not self.field_name in getattr(self.parent.Meta.model._meta, 'nested_fields', []):
                    is_container = False

        cache_vary = str(user)
        cache_result = self.check_cache(value, id, child_model, cache_vary)
        if cache_result:
            return cache_result

        data = super().to_representation(value)
        if is_container:
            data = self.serialize_container(data, id, user, child_model)

        GLOBAL_SERIALIZER_CACHE.set(getattr(child_model._meta, 'label'), id, cache_vary, data)
        return GLOBAL_SERIALIZER_CACHE.get(getattr(child_model._meta, 'label'), id, cache_vary)

    def get_value(self, dictionary):
        try:
            object_list = dictionary['@graph']

            if self.parent.instance is None:
                obj = next(filter(
                    lambda o: not hasattr(o, self.parent.url_field_name) or "./" in o[self.parent.url_field_name],
                    object_list))
            else:
                container_id = Model.container_id(self.parent.instance)
                obj = next(filter(lambda o: container_id.lstrip('/') in o[self.parent.url_field_name], object_list))
            list = super().get_value(obj)
            try:
                list = next(
                    filter(lambda o: list[self.parent.url_field_name] == o[self.parent.url_field_name], object_list))
            except (KeyError, TypeError, StopIteration):
                pass

            try:
                list = list['ldp:contains']
            except (KeyError, TypeError):
                pass

            if list is empty:
                return []

            if isinstance(list, dict):
                list = [list]

            ret = []
            for item in list:
                full_item = None
                try:
                    full_item = next(filter(
                        lambda o: self.parent.url_field_name in o and item[self.parent.url_field_name] == o[
                            self.parent.url_field_name], object_list))
                except StopIteration:
                    pass
                if full_item is None:
                    ret.append(item)
                else:
                    ret.append(full_item)

            return ret
        except KeyError:
            obj = super().get_value(dictionary)
            if isinstance(obj, dict) and self.parent.url_field_name in obj:
                resource_id = obj[self.parent.url_field_name]
                if isinstance(resource_id, str) and resource_id.startswith("_:"):
                    object_list = self.root.initial_data['@graph']
                    obj = [next(filter(lambda o: resource_id in o[self.parent.url_field_name], object_list))]

            return obj


class IdentityFieldMixin:
    def to_internal_value(self, data):
        '''Gives the @id as a representation if present'''
        try:
            return super().to_internal_value(data[self.parent.url_field_name])
        except (KeyError, TypeError):
            return super().to_internal_value(data)



import requests
from django.db import models
from rest_framework import fields


class IdURLField(fields.URLField):

    def to_representation(self, value):
        str = super(IdURLField, self).to_representation(value)
        return {"@id": str}

    def get(self, value):
        url = super(IdURLField, self).to_representation(value)
        datas = requests.get(url).json()
        return datas


class LDPFieldMixin(models.Field):
    """
    Extends Django field to store linked data information.
    """

    def __init__(self, *args, rdf_type=None, related_rdf_type=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.rdf_type = rdf_type
        self.related_rdf_type = related_rdf_type


class BigIntegerField(LDPFieldMixin, models.BigIntegerField):
    pass


class BinaryField(LDPFieldMixin, models.BinaryField):
    pass


class BooleanField(LDPFieldMixin, models.BooleanField):
    pass


class CharField(LDPFieldMixin, models.CharField):
    pass


class DateField(LDPFieldMixin, models.DateField):
    pass


class DateTimeField(LDPFieldMixin, models.DateTimeField):
    pass


class DecimalField(LDPFieldMixin, models.DecimalField):
    pass


class DurationField(LDPFieldMixin, models.DurationField):
    pass


class EmailField(LDPFieldMixin, models.EmailField):
    pass


class FileField(LDPFieldMixin, models.FileField):
    pass


class FilePathField(LDPFieldMixin, models.FilePathField):
    pass


class FloatField(LDPFieldMixin, models.FloatField):
    pass


class GenericIPAddressField(LDPFieldMixin, models.GenericIPAddressField):
    pass


class ImageField(LDPFieldMixin, models.ImageField):
    pass


class IntegerField(LDPFieldMixin, models.IntegerField):
    pass


class JSONField(LDPFieldMixin, models.JSONField):
    pass


class PositiveBigIntegerField(LDPFieldMixin, models.PositiveBigIntegerField):
    pass


class PositiveIntegerField(LDPFieldMixin, models.PositiveIntegerField):
    pass


class PositiveSmallIntegerField(LDPFieldMixin, models.PositiveSmallIntegerField):
    pass


class SlugField(LDPFieldMixin, models.SlugField):
    pass


class SmallAutoField(LDPFieldMixin, models.SmallAutoField):
    pass


class SmallIntegerField(LDPFieldMixin, models.SmallIntegerField):
    pass


class TextField(LDPFieldMixin, models.TextField):
    pass


class TimeField(LDPFieldMixin, models.TimeField):
    pass


class LDPUrlField(LDPFieldMixin, models.URLField):
    pass


class UUIDField(LDPFieldMixin, models.UUIDField):
    pass


class ForeignKey(LDPFieldMixin, models.ForeignKey):
    pass


class ManyToManyField(LDPFieldMixin, models.ManyToManyField):
    pass


class OneToOneField(LDPFieldMixin, models.OneToOneField):
    pass



from django.apps import apps
from django.conf import settings

from djangoldp.views.ldp_api import LDPAPIView

from rest_framework.response import Response


class PublicTypeIndexView(LDPAPIView):
    def get(self, request, *args, **kwargs):
        return self.on_request(request)

    def on_request(self, request):
        # Generate a jsonld response with the context and a @graph containing all the models in the system
        response = {
            '@graph': [{
              "@id": request.build_absolute_uri('publicTypeIndex'),
              "@type": "solid:TypeIndex"
            }]
        }

        # Iterate over all the apps and add their indexes entry-points to the graph
        for app in apps.get_app_configs():
            app_models = [model.__name__ for model in app.get_models()]
            for model_name in app_models:
                model = apps.get_model(app.label, model_name)
                if (model._meta and hasattr(model._meta, 'indexed_fields')):
                    if (hasattr(model._meta, 'rdf_type') and hasattr(model, 'get_container_path')):
                        response['@graph'].append({
                            "@type": "solid:TypeIndexRegistration",
                            "solid:forClass": "idx:Index",
                            '@id': request.build_absolute_uri(f"publicTypeIndex#indexes-{model.get_container_path()[1:-1]}"),
                            "solid:instance": request.build_absolute_uri('/indexes' + model.get_container_path() + 'index')
                        })

        # Iterate over all the apps in the system and their models to add them to the @graph
        for app in apps.get_app_configs():
            app_models = [model.__name__ for model in app.get_models()]
            for model_name in app_models:
                model = apps.get_model(app.label, model_name)
                if (model._meta ):
                    if (hasattr(model._meta, 'rdf_type') and hasattr(model, 'get_container_path')):
                        response['@graph'].append({
                            "@id": request.build_absolute_uri(f"publicTypeIndex#{model.get_container_path()[1:-1]}"),
                            "@type": "solid:TypeIndexRegistration",
                            "solid:forClass": model._meta.rdf_type,
                            "solid:instanceContainer": request.build_absolute_uri(model.get_container_path())
                        })

        return Response(response,
                        content_type='application/ld+json',
                        headers={
                          'Access-Control-Allow-Origin': '*',
                          'Cache-Control': 'public, max-age=3600',
                        })



import unittest
"""
Comprehensive tests for Phase 2 W3C LDP compliance:
- Prefer headers (RFC 7240)
- OPTIONS method support

Tests compliance with:
- RFC 7240 (Prefer Header for HTTP)
- W3C LDP specification
- HTTP OPTIONS method best practices
"""

import json
from django.contrib.auth import get_user_model
from rest_framework.test import APIClient, APIRequestFactory, APITestCase

from djangoldp.serializers import GLOBAL_SERIALIZER_CACHE
from djangoldp.tests.models import Post


class TestPreferHeaders(APITestCase):
    """Test Prefer header support for minimal/representation preferences."""

    def setUp(self):
        """Set up test fixtures."""
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(
            username='testuser',
            email='test@example.com',
            password='testpass'
        )
        GLOBAL_SERIALIZER_CACHE.reset()

    def tearDown(self):
        """Clean up after tests."""
        GLOBAL_SERIALIZER_CACHE.reset()

    # ===== POST (Create) with Prefer Headers =====

    def test_post_prefer_minimal(self):
        """Test POST with Prefer: return=minimal returns 204 with Location."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return=minimal'
        )

        # Should return 204 No Content
        self.assertEqual(response.status_code, 204)

        # Should have Location header
        self.assertIn('Location', response)
        self.assertIsNotNone(response['Location'])
        self.assertTrue(len(response['Location']) > 0)

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should have Preference-Applied header
        self.assertIn('Preference-Applied', response)
        self.assertEqual(response['Preference-Applied'], 'return=minimal')

        # Should have no body content
        self.assertEqual(len(response.content), 0)

    def test_post_prefer_representation(self):
        """Test POST with Prefer: return=representation returns 201 with body."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return=representation'
        )

        # Should return 201 Created
        self.assertEqual(response.status_code, 201)

        # Should have body content
        self.assertIn('content', response.data)
        self.assertEqual(response.data['content'], 'new post content')

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should have Preference-Applied header
        self.assertIn('Preference-Applied', response)
        self.assertEqual(response['Preference-Applied'], 'return=representation')

    def test_post_no_prefer_header(self):
        """Test POST without Prefer header defaults to full representation."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json'
        )

        # Should return 201 Created (default behavior)
        self.assertEqual(response.status_code, 201)

        # Should have body content
        self.assertIn('content', response.data)
        self.assertEqual(response.data['content'], 'new post content')

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should NOT have Preference-Applied header (no preference specified)
        self.assertNotIn('Preference-Applied', response)

    def test_post_prefer_both_minimal_and_representation(self):
        """Test POST with both preferences - representation should take precedence."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return=minimal, return=representation'
        )

        # When both are present, return=representation takes precedence
        # (implementation returns representation when both are set)
        self.assertEqual(response.status_code, 201)
        self.assertIn('content', response.data)

    def test_post_prefer_case_insensitive(self):
        """Test that Prefer header parsing is case-insensitive."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}

        # Test uppercase
        response1 = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='RETURN=MINIMAL'
        )
        self.assertEqual(response1.status_code, 204)

        # Test mixed case
        response2 = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='ReTuRn=MiNiMaL'
        )
        self.assertEqual(response2.status_code, 204)

    # ===== PUT (Update) with Prefer Headers =====

    def test_put_prefer_minimal(self):
        """Test PUT with Prefer: return=minimal returns 204 with Location."""
        post = Post.objects.create(content="original content")

        # Get current ETag
        get_response = self.client.get(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        etag = get_response['ETag']

        # Update with Prefer: return=minimal
        data = {'https://cdn.startinblox.com/owl#content': 'updated content'}
        response = self.client.put(
            f'/posts/{post.pk}/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_IF_MATCH=etag,
            HTTP_PREFER='return=minimal'
        )

        # Should return 204 No Content
        self.assertEqual(response.status_code, 204)

        # Should have Location header
        self.assertIn('Location', response)
        self.assertIsNotNone(response['Location'])

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should have Preference-Applied header
        self.assertIn('Preference-Applied', response)
        self.assertEqual(response['Preference-Applied'], 'return=minimal')

        # Should have no body content
        self.assertEqual(len(response.content), 0)

        # Verify the update actually happened
        post.refresh_from_db()
        self.assertEqual(post.content, 'updated content')

    def test_put_prefer_representation(self):
        """Test PUT with Prefer: return=representation returns 200 with body."""
        post = Post.objects.create(content="original content")

        # Get current ETag
        get_response = self.client.get(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        etag = get_response['ETag']

        # Update with Prefer: return=representation
        data = {'https://cdn.startinblox.com/owl#content': 'updated content'}
        response = self.client.put(
            f'/posts/{post.pk}/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_IF_MATCH=etag,
            HTTP_PREFER='return=representation'
        )

        # Should return 200 OK
        self.assertEqual(response.status_code, 200)

        # Should have body content
        self.assertIn('content', response.data)
        self.assertEqual(response.data['content'], 'updated content')

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should have Preference-Applied header
        self.assertIn('Preference-Applied', response)
        self.assertEqual(response['Preference-Applied'], 'return=representation')

    def test_put_no_prefer_header(self):
        """Test PUT without Prefer header defaults to full representation."""
        post = Post.objects.create(content="original content")

        # Get current ETag
        get_response = self.client.get(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        etag = get_response['ETag']

        # Update without Prefer header
        data = {'https://cdn.startinblox.com/owl#content': 'updated content'}
        response = self.client.put(
            f'/posts/{post.pk}/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_IF_MATCH=etag
        )

        # Should return 200 OK (default behavior)
        self.assertEqual(response.status_code, 200)

        # Should have body content
        self.assertIn('content', response.data)
        self.assertEqual(response.data['content'], 'updated content')

        # Should have ETag header
        self.assertIn('ETag', response)

        # Should NOT have Preference-Applied header
        self.assertNotIn('Preference-Applied', response)

    # ===== PATCH (Partial Update) with Prefer Headers =====

    def test_patch_prefer_minimal(self):
        """Test PATCH with Prefer: return=minimal returns 204 with Location."""
        post = Post.objects.create(content="original content")

        # Get current ETag
        get_response = self.client.get(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        etag = get_response['ETag']

        # Patch with Prefer: return=minimal
        data = {'https://cdn.startinblox.com/owl#content': 'patched content'}
        response = self.client.patch(
            f'/posts/{post.pk}/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_IF_MATCH=etag,
            HTTP_PREFER='return=minimal'
        )

        # Should return 204 No Content
        self.assertEqual(response.status_code, 204)

        # Should have Preference-Applied header
        self.assertIn('Preference-Applied', response)
        self.assertEqual(response['Preference-Applied'], 'return=minimal')

        # Verify the update actually happened
        post.refresh_from_db()
        self.assertEqual(post.content, 'patched content')

    # ===== Prefer Header Edge Cases =====

    def test_prefer_header_with_additional_parameters(self):
        """Test Prefer header with additional parameters."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return=minimal; wait=100'
        )

        # Should still recognize return=minimal
        self.assertEqual(response.status_code, 204)
        self.assertIn('Preference-Applied', response)

    def test_prefer_minimal_with_spaces(self):
        """Test Prefer header with various spacing."""
        data = {'https://cdn.startinblox.com/owl#content': 'new post content'}

        # Test with spaces
        response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return = minimal'
        )

        # Should handle spacing variations
        # Note: Our implementation uses 'return=minimal' in prefer_header.lower()
        # which won't match 'return = minimal', so this should return 201
        self.assertIn(response.status_code, [201, 204])


class TestOPTIONSMethod(APITestCase):
    """Test OPTIONS method support with proper LDP headers."""

    def setUp(self):
        """Set up test fixtures."""
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(
            username='testuser',
            email='test@example.com',
            password='testpass'
        )
        GLOBAL_SERIALIZER_CACHE.reset()

    def tearDown(self):
        """Clean up after tests."""
        GLOBAL_SERIALIZER_CACHE.reset()

    # ===== OPTIONS on Container (List) Views =====

    def test_options_on_container(self):
        """Test OPTIONS on container returns proper headers."""
        response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        # Should return 200 OK
        self.assertEqual(response.status_code, 200)

        # Should have Allow header
        self.assertIn('Allow', response)
        allow_methods = [m.strip() for m in response['Allow'].split(',')]
        self.assertIn('GET', allow_methods)
        self.assertIn('POST', allow_methods)
        self.assertIn('HEAD', allow_methods)
        self.assertIn('OPTIONS', allow_methods)
        # Container should not have PUT, PATCH, DELETE
        self.assertNotIn('PUT', allow_methods)
        self.assertNotIn('PATCH', allow_methods)
        self.assertNotIn('DELETE', allow_methods)

    def test_options_container_accept_post(self):
        """Test OPTIONS on container includes Accept-Post header."""
        response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        self.assertIn('Accept-Post', response)
        accept_post = response['Accept-Post']
        self.assertIn('application/ld+json', accept_post)
        self.assertIn('text/turtle', accept_post)

    def test_options_container_link_headers(self):
        """Test OPTIONS on container includes proper LDP Link headers."""
        response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        self.assertIn('Link', response)
        link_header = response['Link']

        # Should include Container types
        self.assertIn('http://www.w3.org/ns/ldp#Resource', link_header)
        self.assertIn('http://www.w3.org/ns/ldp#RDFSource', link_header)
        self.assertIn('http://www.w3.org/ns/ldp#Container', link_header)
        self.assertIn('http://www.w3.org/ns/ldp#BasicContainer', link_header)

    def test_options_container_no_accept_patch(self):
        """Test OPTIONS on container does not include Accept-Patch."""
        response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        # Containers should not have Accept-Patch
        self.assertNotIn('Accept-Patch', response)

    # ===== OPTIONS on Detail (Resource) Views =====

    def test_options_on_detail(self):
        """Test OPTIONS on detail view returns proper headers."""
        post = Post.objects.create(content="test content")
        response = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )

        # Should return 200 OK
        self.assertEqual(response.status_code, 200)

        # Should have Allow header
        self.assertIn('Allow', response)
        allow_methods = [m.strip() for m in response['Allow'].split(',')]
        self.assertIn('GET', allow_methods)
        self.assertIn('PUT', allow_methods)
        self.assertIn('PATCH', allow_methods)
        self.assertIn('DELETE', allow_methods)
        self.assertIn('HEAD', allow_methods)
        self.assertIn('OPTIONS', allow_methods)
        # Detail should not have POST
        self.assertNotIn('POST', allow_methods)

    def test_options_detail_accept_patch(self):
        """Test OPTIONS on detail view includes Accept-Patch header."""
        post = Post.objects.create(content="test content")
        response = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )

        self.assertIn('Accept-Patch', response)
        accept_patch = response['Accept-Patch']
        self.assertIn('application/ld+json', accept_patch)
        self.assertIn('text/turtle', accept_patch)

    def test_options_detail_link_headers(self):
        """Test OPTIONS on detail view includes proper LDP Link headers."""
        post = Post.objects.create(content="test content")
        response = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )

        self.assertIn('Link', response)
        link_header = response['Link']

        # Should include Resource types but not Container types
        self.assertIn('http://www.w3.org/ns/ldp#Resource', link_header)
        self.assertIn('http://www.w3.org/ns/ldp#RDFSource', link_header)
        # Should NOT include Container types
        self.assertNotIn('http://www.w3.org/ns/ldp#Container', link_header)
        self.assertNotIn('http://www.w3.org/ns/ldp#BasicContainer', link_header)

    def test_options_detail_no_accept_post(self):
        """Test OPTIONS on detail view does not include Accept-Post."""
        post = Post.objects.create(content="test content")
        response = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )

        # Detail views should not have Accept-Post
        self.assertNotIn('Accept-Post', response)

    # ===== OPTIONS Response Body =====

    def test_options_empty_body(self):
        """Test OPTIONS returns empty body."""
        response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        # Should have empty body
        self.assertEqual(len(response.content), 0)

    # ===== Integration: OPTIONS with other methods =====

    def test_options_reflects_actual_capabilities(self):
        """Test that OPTIONS Allow header matches actual supported methods."""
        post = Post.objects.create(content="test content")

        # Get OPTIONS
        options_response = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        allow_methods = [m.strip() for m in options_response['Allow'].split(',')]

        # Test that each allowed method actually works
        if 'GET' in allow_methods:
            get_response = self.client.get(f'/posts/{post.pk}/')
            self.assertNotEqual(get_response.status_code, 405)

        if 'PUT' in allow_methods:
            # Need to get ETag first for conditional update
            get_response = self.client.get(f'/posts/{post.pk}/')
            etag = get_response['ETag']
            data = {'https://cdn.startinblox.com/owl#content': 'updated'}
            put_response = self.client.put(
                f'/posts/{post.pk}/',
                data=json.dumps(data),
                content_type='application/ld+json',
                HTTP_IF_MATCH=etag
            )
            self.assertNotEqual(put_response.status_code, 405)

        if 'DELETE' in allow_methods:
            delete_response = self.client.delete(f'/posts/{post.pk}/')
            self.assertNotEqual(delete_response.status_code, 405)


class TestPreferAndOPTIONSIntegration(APITestCase):
    """Test integration between Prefer headers and OPTIONS method."""

    def setUp(self):
        """Set up test fixtures."""
        self.client = APIClient()
        GLOBAL_SERIALIZER_CACHE.reset()

    def tearDown(self):
        """Clean up after tests."""
        GLOBAL_SERIALIZER_CACHE.reset()

    def test_options_then_post_with_prefer(self):
        """Test workflow: OPTIONS to discover capabilities, then POST with Prefer."""
        # First, use OPTIONS to discover what's supported
        options_response = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )

        # Verify POST is supported
        allow_methods = [m.strip() for m in options_response['Allow'].split(',')]
        self.assertIn('POST', allow_methods)

        # Verify Accept-Post includes our content type
        self.assertIn('Accept-Post', options_response)
        self.assertIn('application/ld+json', options_response['Accept-Post'])

        # Now POST with Prefer header
        data = {'https://cdn.startinblox.com/owl#content': 'new post'}
        post_response = self.client.post(
            '/posts/',
            data=json.dumps(data),
            content_type='application/ld+json',
            HTTP_PREFER='return=minimal'
        )

        self.assertEqual(post_response.status_code, 204)
        self.assertIn('Preference-Applied', post_response)

    def test_prefer_header_in_allowed_methods(self):
        """Test that methods that support Prefer are in Allow header."""
        # Container OPTIONS
        container_options = self.client.options(
            '/posts/',
            content_type='application/ld+json'
        )
        container_methods = [m.strip() for m in container_options['Allow'].split(',')]

        # POST should be allowed (supports Prefer)
        self.assertIn('POST', container_methods)

        # Detail OPTIONS
        post = Post.objects.create(content="test")
        detail_options = self.client.options(
            f'/posts/{post.pk}/',
            content_type='application/ld+json'
        )
        detail_methods = [m.strip() for m in detail_options['Allow'].split(',')]

        # PUT and PATCH should be allowed (support Prefer)
        self.assertIn('PUT', detail_methods)
        self.assertIn('PATCH', detail_methods)



# Generated migration for adding timestamp fields to all LDP models
# These fields support HTTP caching headers (Last-Modified, If-Modified-Since)

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0020_rename_sitesettings_sitesetting'),
    ]

    operations = [
        # Activity already has created_at from migration 0006, only add updated_at
        migrations.AddField(
            model_name='activity',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        # Add created_at field to Follower model
        migrations.AddField(
            model_name='follower',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        # Add updated_at field to Follower model
        migrations.AddField(
            model_name='follower',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        # Add created_at field to LDPSource model
        migrations.AddField(
            model_name='ldpsource',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        # Add updated_at field to LDPSource model
        migrations.AddField(
            model_name='ldpsource',
            name='updated_at',
            field=models.DateTimeField(auto_now=True),
        ),
        # Note: SiteSetting is not an LDP model (inherits from models.Model, not djangoldp.models.Model)
        # so it does not get timestamp fields
    ]



from django.conf import settings

# defaults for various DjangoLDP settings (see documentation)
MAX_RECORDS_SERIALIZER_CACHE = getattr(settings, 'MAX_RECORDS_SERIALIZER_CACHE', 10000)


class InMemoryCache:
    def __init__(self):
        self.cache = {}

    def reset(self):
        self.cache = {}

    def has(self, cache_key, container_urlid=None, vary=None):
        return cache_key in self.cache and \
               (container_urlid is None or container_urlid in self.cache[cache_key]) and \
               (vary is None or vary in self.cache[cache_key][container_urlid])

    def get(self, cache_key, container_urlid, vary):
        if self.has(cache_key, container_urlid, vary):
            return self.cache[cache_key][container_urlid][vary]['value']
        else:
            return None

    def set(self, cache_key, container_urlid, vary, value):
        if len(self.cache.keys()) > MAX_RECORDS_SERIALIZER_CACHE:
            self.reset()

        if cache_key not in self.cache:
            self.cache[cache_key] = {}
        if container_urlid not in self.cache[cache_key]:
            self.cache[cache_key][container_urlid] = {}
        self.cache[cache_key][container_urlid][vary] = {'value': value}

    def invalidate(self, cache_key, container_urlid=None, vary=None):
        # can clear cache_key -> container_urlid -> vary, cache_key -> container_urlid or cache_key
        if container_urlid is not None:
            if vary is not None:
                self.cache[cache_key][container_urlid].pop(vary, None)
            else:
                self.cache[cache_key].pop(container_urlid, None)
        else:
            self.cache.pop(cache_key, None)


GLOBAL_SERIALIZER_CACHE = InMemoryCache()



from django.apps import apps
from django.conf import settings

from djangoldp.views.ldp_api import LDPAPIView

from rest_framework.response import Response
from rest_framework import status

import logging

logger = logging.getLogger(__name__)

class InstanceRootContainerView(LDPAPIView):
    def get(self, request, *args, **kwargs):
        return self.on_request(request)

    def on_request(self, request):
        try:
          # Generate a jsonld response with the context and a @graph containing all the models in the system
          response = {
              '@id': request.build_absolute_uri(),
              '@type': 'ldp:Container',
              'ldp:contains': []
          }

          # Iterate over all the apps in the system and their models to add them to the @graph
          for app in apps.get_app_configs():
              app_models = [model.__name__ for model in app.get_models()]
              for model_name in app_models:
                  model = apps.get_model(app.label, model_name)
                  if (model._meta ):
                      if (hasattr(model._meta, 'rdf_type') and hasattr(model, 'get_container_path')):
                          response['ldp:contains'].append({
                              "@id": request.build_absolute_uri(model.get_container_path()),
                              "@type": "ldp:Container"
                          })
          response = Response(response,
                          content_type='application/ld+json',
                          headers={
                            'Access-Control-Allow-Origin': '*',
                            'Cache-Control': 'public, max-age=3600',
                          })

          return response
        except Exception as e:
            logger.exception("Error building LDP container response")
            return Response(
                {'error': 'Internal server error'},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR
            )




# Generated by Django 4.2 on 2025-06-17 14:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djangoldp', '0019_sitesettings'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='SiteSettings',
            new_name='SiteSetting',
        ),
    ]



"""
ETag generation utilities for DjangoLDP resources and containers.

Implements W3C LDP compliance for entity tag support.
"""

import hashlib
import json
from typing import Optional, Any
from django.core.exceptions import FieldDoesNotExist
from django.db.models import QuerySet

# Constant for timestamp conversion
TIMESTAMP_TO_MICROSECONDS = 1000000


def normalize_etag(etag_str):
    """
    Normalize ETag for comparison.

    Args:
        etag_str: ETag string from header

    Returns:
        tuple: (is_weak: bool, etag_value: str)
    """
    if not etag_str:
        return False, ""

    etag_str = etag_str.strip()
    is_weak = etag_str.startswith('W/')
    if is_weak:
        etag_str = etag_str[2:]
    return is_weak, etag_str.strip('"')


def generate_etag(obj: Any, serialized_data: Optional[dict] = None) -> str:
    """
    Generate weak ETag for a model instance.
    Uses object's updated_at if available, otherwise hashes serialized content.

    Args:
        obj: Django model instance
        serialized_data: Optional serialized representation of the object

    Returns:
        str: Weak ETag value in format W/"hash"
    """
    # Try timestamp-based ETag first (more efficient)
    if hasattr(obj, 'updated_at') and obj.updated_at:
        try:
            timestamp = obj.updated_at.timestamp()
            timestamp_micro = int(timestamp * TIMESTAMP_TO_MICROSECONDS)
            return f'W/"{timestamp_micro}"'
        except (AttributeError, ValueError, TypeError):
            # Fall through to content-based ETag
            pass

    # Fall back to content-based ETag
    try:
        if serialized_data:
            content = json.dumps(serialized_data, sort_keys=True)
        else:
            # Use pk and model name as minimal identifier
            content = f"{obj.__class__.__name__}:{obj.pk}"
    except (TypeError, ValueError):
        # If serialization fails, use minimal identifier
        content = f"{obj.__class__.__name__}:{obj.pk}"

    # Use SHA256 instead of MD5 for better security
    hash_value = hashlib.sha256(content.encode()).hexdigest()[:32]  # Truncate to 32 chars
    return f'W/"{hash_value}"'  # Weak ETag


def generate_container_etag(queryset: QuerySet, count: int, page_number: Optional[int] = None) -> str:
    """
    Generate weak ETag for a container.
    Based on count, latest modification time, and optional page number.

    Args:
        queryset: Django QuerySet for the container
        count: Number of items in the container (total count, not page size)
        page_number: Optional page number for paginated containers

    Returns:
        str: Weak ETag value for the container in format W/"hash"
    """
    latest_updated = None

    # Check if model has updated_at field using FieldDoesNotExist
    try:
        queryset.model._meta.get_field('updated_at')
        # Field exists, get the latest timestamp
        latest_timestamp = queryset.order_by('-updated_at').values_list('updated_at', flat=True).first()
        if latest_timestamp:
            try:
                latest_updated = latest_timestamp.timestamp()
            except (AttributeError, ValueError, TypeError):
                pass
    except FieldDoesNotExist:
        # Field doesn't exist, continue with fallback
        pass

    if latest_updated:
        timestamp_micro = int(latest_updated * TIMESTAMP_TO_MICROSECONDS)
        etag_content = f"{count}:{timestamp_micro}"
    else:
        # Fallback: use count and deterministic hash of query
        query_str = str(queryset.query)
        # Use SHA256 for deterministic hashing instead of built-in hash()
        query_hash = hashlib.sha256(query_str.encode()).hexdigest()[:16]
        etag_content = f"{count}:{query_hash}"

    # Include page number in ETag if paginated
    if page_number is not None:
        etag_content = f"{etag_content}:page{page_number}"

    # Use SHA256 for final hash
    hash_value = hashlib.sha256(etag_content.encode()).hexdigest()[:32]  # Truncate to 32 chars
    return f'W/"{hash_value}"'
